/*
 * ifd_seekfree - seek free block chunk on INDEX file.
 *      The chunk have to no break.  Moreover its length is longer
 *		or equal than the target object.
 *
 * return:
 *  success     0
 *  error       <0
 *
 */
int
ifd_seekfree(int sfd, bucket_t *bk, skcache_entry_t *e, int *rpos)
{
    int ret;
    int c;
    off_t pos, dst;
    int fnum;
    int ipos, npos;     /* position by blocks */
    int nr;
    skcache_entry_t *tmp_entry;
    int match = 0;
    int try;
    off_t vpos;
    time_t diff;
    skcache_entry_t chunk[HARD_NINDEX_DREAD];
    int i;
    int nwid, wid;
    int first;
    int want;
    int xid; /* loop id */
    int sig_pos;

    ret = 0;
    *rpos = (off_t) -1;
    xid = rand()%1000;  /* make temporary loop id */
    
    if(e==NULL) {
    }

    fnum = FDPOOL_POS_E(e);
    sig_pos = DSIG_POS(e->h1, e->h2, e->h3);
    want = NBCEIL(e->totallength, OFILE_BLKSIZE);

    npos = ipos = ofile_offset(e);
    pos = (off_t)npos * ESIZE;
    POSCHECKI(pos);
    if(pos<0 || npos<0) {
        Error("psf: 1 Ignore position block# %d %lldB; e h3 %d h4 %d; slot %d",
            npos, (long long)pos, e->h3, e->h4, OFILE_NBLK_FILE);
        ret = XSK_BODY_EPOS;
        goto out;
    }

#if 0
    if(npos+want-1<=OFILE_NBLK_FILE-1) {
        goto not_found;
    }
#endif

#if 0
    if(npos+want-1>OFILE_NBLK_FILE-1) {
        goto not_found;
    }
#endif

#if 0
    Trace("psf: 0 [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_BLKSIZE);
#endif

#if 0
 if(DFC) {
Trace("psf: 2 sfd# %d, [%03d:%03d:%03d:%04d] x %6ld => pos %8d/%d %12lldB (%4ldMB+%7ld)\n",
    sfd,
    e->h1, e->h2, e->h3, e->h4, OFILE_BLKSIZE,
    npos, OFILE_NBLK_FILE,
    (long long)pos, (long)dM(pos), (long)mM(pos) );
 }
#endif

 if(DFC) {
Trace("psf: 2 dir# %d, sfd# %d, %d:%d:%d:%d from pos %d with %d blocks ; count %d\n",
    sig_pos, sfd, e->h1, e->h2, e->h3, e->h4, npos, want, dir_count[sig_pos]);
 }



#if 1
    /*
     * bitmapping is enable.
     * hook position scanning.
     */
    if(bitmapping_flag) {
        if(dir_count[sig_pos]<0) {
#if 0
            Trace("psf: %3d don't ref. dir_bit\n", xid);
#endif

            /* XXX, currently this disk seeking routine does not support
               multi block seeking */

#if 0
            if(want>1) {
                goto not_found;
            }
#endif
#if 0
            if(want>1) {
                Trace("  oid# %d, want %d\n", e->oid, want);
            }
#endif

        }
        else {
            int tmp_pos;

            tmp_pos = _psf_bit_firstmatch(sig_pos, want, npos, DFC);
#if 0
            Trace("result by first-match %d\n", tmp_pos);
#endif

            if(tmp_pos<0) {
                goto not_found;
            }
            else {
                npos = tmp_pos;
                /* XXX */
                pos = (off_t)npos * ESIZE;
                goto found_by_membits;
            }
        }
    }
    else {
    }
#endif

    POSCHECKI(pos);

    if(pos<(off_t)0L) {
        Error("psf: ignore position block# %d %lldB before seek()", 
            npos, (long long)pos);

        ret = XSK_BODY_EPOS;
        goto out;
    }

#if 0
    if(POSVERIFY(sfd, ifd_pos[sig_pos])) {
        ifd_pos[sig_pos] = -1;
    }
#endif

    c = 0;
try_seek:
    c++;
    dst = Iseek(sig_pos, sfd, pos);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
Error("psf: cannot seek %d %lldB on index file fd# %d :%d (%d)",
            npos, (long long)pos, sfd, c, errno);

        ret = XSK_BODY_EPOS;
        goto out;
    }
    vpos = dst;
    POSVERIFY(sfd, vpos);
    POSCHECKI(dst);

    try = OFILE_NBLK_SEEK;
    match = 0;
    first = -1;
    do {
#if 0
        if(npos+want-1>=OFILE_NBLK_FILE) {
            match = 0;
            break;
        }
#endif

        pos = (off_t)npos * ESIZE;
        nwid = MIN(npos+ MIN(nindex_dread, OFILE_NBLK_SEEK),
                    OFILE_NBLK_FILE) - npos;
        wid = ESIZE*nwid;

        /* block read */

 if(DFC) {
Trace("psf: 4 sfd# %d, pos %d %lld, addr %p, wid %d %d => %d/%d %lld/%lld\n",
    sfd, npos, (long long)pos, chunk, nwid, wid, npos+nwid, OFILE_NBLK_FILE,
    (long long)(pos+wid), (long long)(OFILE_NBLK_FILE*ESIZE) );
 }

        c = 0;
try_read:
        c++;

        nr = Iread(sig_pos, sfd, chunk, wid);
        if(nr<0) {
            if(errno==EINTR && c < DISK_RETRY) {
                goto try_read;
            }
            Error("psf: cannot read fd# %d: %d (%d)",
                sfd, c, errno);
            break;
        }
        vpos += nr;
        POSVERIFY(sfd, vpos);
        nwid = nr/ESIZE;

#if 0
        for(i=0;i<nwid && try && match<want; i++) 
#endif
        for(i=0;i<nwid && try ; i++) 
        {
            pos = (off_t)npos * ESIZE;
            tmp_entry = &chunk[i];

            if(npos==0) {
                match = 0;
                first = -1;
            }

            /*
             * ignore magic, empty or very older is counted as target
             */
            if((tmp_entry->magic[0]!=MAGIC_0 || tmp_entry->magic[1]!=MAGIC_1)
                || tmp_entry->oid == 0 || tmp_entry->issue_time < min_date) {

                match++;
                if(first<0) {
                    first = npos;
                }
            }
            else {
                match = 0;
                first = -1;
            }

#ifdef TRACE_INDEX
#endif
 if(DFC) {
        diff = tmp_entry->issue_time - min_date;

#if 0
Trace("  :f dir# %d, fd# %d, try %3d: pos %4d %12lldB : vpos %12lldB : magic(hex) %02x %02x, oid# %d, diff %d; match %d, first %d\n",
        sig_pos, sfd, try, npos, (long long)pos, (long long) vpos,
        (unsigned char)tmp_entry->magic[0], (unsigned char)tmp_entry->magic[1],
        tmp_entry->oid, diff, match, first
        );
#endif

Trace("  :f dir# %d, ? pos %4d %4d, try %3d: %d: %02x %02x, oid# %d, hash %d:%d:%d:%d diff %d\n",
        sig_pos, npos, npos, try,
        match,
        (unsigned char)tmp_entry->magic[0], (unsigned char)tmp_entry->magic[1],
        tmp_entry->oid, 
        tmp_entry->h1, tmp_entry->h2, tmp_entry->h3, tmp_entry->h4,
        diff
        );

 }

#if 1
            if(match>=want) {
                break;
            }

            npos = (npos+1)% OFILE_NBLK_FILE;
            try--;

            if(npos==0) {
                break;
            }
#else

            if(npos<OFILE_NBLK_FILE-1) {
                npos++;
            }
#endif

        }
        if(match>=want) {
            break;
        }
        
        if(npos==0) {
#ifdef TRACE_INDEX
            Trace("  : rewind sfd# %d,\n", sfd);
#endif
            c = 0;
try_rewind:
            c++;
            dst = Iseek(sig_pos, sfd, 0);
            if(dst<(off_t)0) {
                if(errno==EINTR && c<DISK_RETRY) {
                    goto try_rewind;
                }
    Error("psf: cannot rewind %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);
            }
            vpos = dst;
            POSVERIFY(sfd, vpos);
            POSCHECKI(dst);
        
            match = 0;
            first = -1;
        }
    } while (match==0 && try>0);


    /* 
     * find by disk seeking
     *  - move file pointer to first block on INDEX file
     *  - set return code and position
     */

#if 0
    if(match>=want && first>=0)
#endif
    if(match && first>=0)
    {
#if 0
    Trace("psf: %3d found in disk; block# %d + %d\n", xid, npos, want);
#endif
        POSCHECKI(ifd_pos[sig_pos]);

        c = 0;
try_back:
        c++;
        npos = first;
        pos = first*ESIZE;
        dst = Iseek(sig_pos, sfd, pos);
        if(dst<(off_t)0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_back;
            }
    Error("psf: cannot back %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);

            ret = XSK_BODY_EROLL;
            *rpos = npos;
            goto out;
        }
        vpos = dst;
        POSVERIFY(sfd, vpos);
        POSCHECKI(dst);

#if 0
        ret = npos;
#endif
        ret = 0;
        *rpos = npos;

        goto found;
    }
    else {
#if 0
Trace("psf: 1 sfd# %d, not found (want %d), [%d:%d:%d:%d] fnum %d, npos %d\n",
        sfd, want, e->h1, e->h2, e->h3, e->h4, FDPOOL_POS_E(e), ofile_offset(e));
#endif
        goto not_found;
    }

    goto out;


    /* 
     * find by bitmask on memory seeking
     *  - move file pointer to first block on INDEX file
     *  - set return code and position
     */

found_by_membits:
#if 0
    Trace("psf: %3d found in mem; block# %d + %d\n", xid, first, want);
#endif
    pos = (off_t)npos * ESIZE;

    c = 0;
try_pos3:
    c++;
    dst = Iseek(sig_pos, sfd, pos);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_pos3;
        }
Error("psf: cannot seek %lld on index file fd# %d :%d (%d)",
    (long long)pos, sfd, c, errno);

        ret = XSK_BODY_EPOS;
        *rpos = npos;
        goto out;
    }
    vpos = dst;
    POSVERIFY(sfd, vpos);
    POSCHECKI(dst);

    ret = 0;
    *rpos = npos;

    goto found;

found:
 if(DFC) {
    Trace(
"psf: 5 dir# %d, fd# %d, want %d, hash %d:%d:%d:%d range [%d,%d) ; return %d found [%d..%d)\n",
    sig_pos, sfd, want, e->h1, e->h2, e->h3, e->h4,
    ofile_offset(e), (ofile_offset(e)+OFILE_NBLK_SEEK)%OFILE_NBLK_FILE,
    ret, *rpos, (*rpos+want)%OFILE_NBLK_FILE);
 }
    goto out;

not_found:
    dir_lack[fnum]++;
    *rpos = npos;
    ret = XSK_INDEX_NOTFOUND;

 if(DFC) {
    Trace(
"psf: 6 dir# %d, fd# %d, want %d, hash %d:%d:%d:%d range [%d,%d) ; return %d not-found [%d..%d)\n",
    sig_pos, sfd, want, e->h1, e->h2, e->h3, e->h4,
    ofile_offset(e), (ofile_offset(e)+OFILE_NBLK_SEEK)%OFILE_NBLK_FILE,
    ret, *rpos, (*rpos+want)%OFILE_NBLK_FILE);
 }

    goto out;


out:

#if 0
    /* debug for edge of array */

    if(ipos >= OFILE_NBLK_FILE - OFILE_NBLK_OBJ - 2) {
        Trace(
"psf: dir# %d, fd# %d, ; scan %d..%d, want %d => out %d..%d (%d) with rcode %d\n",
        sig_pos, sfd,
            ipos, ipos+OFILE_NBLK_SEEK, want,
            *rpos, *rpos+want, first, ret);
    }

#endif

#if 0
#endif
#ifdef TRACE_INDEX
    pos = (off_t)*rpos * ESIZE;
Trace(
"psf: 7 [%03d:%03d:%03d:%04d] sfd# %d, try %d, match %d, pos %d %12lld <%12lld>, return %d\n",
    e->h1, e->h2, e->h3, e->h4, sfd, OFILE_NBLK_SEEK - try, match,
    *rpos, (long long) pos, (long long) vpos, ret);
#endif

    return ret;
}
