#!/usr/bin/perl

#
# Perl/Tk script
#

use Tk;

require 'getopts.pl';


&Getopts('hf:');

if($opt_f) {
	$cfgname = $opt_f;
}
else {
	$cfgname = '';
}

if($opt_h) {
	$pgn = $0;
	$pgn =~ s#^.+/##;	# basename

    print <<EOM;
$pgn - make fsmapping configuratino for KOTETU 1.5 and later
    by k-chinen\@is.aist-nara.ac.jp, 2001.

usage:
    $pgn [options] <the number of filesystem> <the number of buckets>

option:
    -h          print this message
	-f file		read file as initial values

example:
	$pgn -f kotetu.conf 28 613

EOM
    exit 0;
}

$nfs = $ARGV[0];
$nbk = $ARGV[1];

if($nfs<=0) {
    $nfs = 8;
}
if($nbk<=0) {
    $nbk = 613;
}

$cwid = 500;
$cht  = 400;
$tht  = 20;

#$zx = $cwid/12;
$zy = $cht - $tht*2; 

#$fwid = ($cwid-2*$zx)/($nfs+2);
$fwid = $cwid/($nfs+2);
$zx = $fwid;

$fnht = $nbk/$nfs * 2;
$fht = int(($cht - $tht)/$fnht);

print "# fwid $fwid, fht $fht\n";

$cnbk = 0;


$mw = MainWindow->new;

$mw->title("mkfsmap - make fsmap for KOTETU");

$msgframe = $mw->Frame(-relief => 'groove', -bd => 2)
    ->pack(-side => 'bottom', -fill => "x");

$mbox = $msgframe->Label(
    -text => "Botton-1: up // Button-3: down // Botton-2: flat")
    ->pack(-side => 'left', -fill => "x");

$cmdframe = $mw->Frame(-relief => 'groove', -bd => 2)
    ->pack(-side => 'bottom', -fill => "x");

$quit = $cmdframe->Button(-text => "Quit", -command => \&exit)
    ->pack(-side => 'left');
$print = $cmdframe->Button(-text => "Print", -command => \&fsprint)
    ->pack(-side => 'left');
$print = $cmdframe->Button(-text => "Reset", -command => \&reset)
    ->pack(-side => 'left');

$noteframe = $mw->Frame(-relief => 'groove', -bd => 2)
    ->pack(-side => 'top', -fill => "x");

$note0 = $noteframe->Label(-text => "The number of buckets:")
    ->pack(-side => 'left', -fill => "x");
$note1 = $noteframe->Label(-text => "given ")
    ->pack(-side => 'left', -fill => "x");
$note2 = $noteframe->Label(-text => "$nbk")
    ->pack(-side => 'left', -fill => "x");
$note3 = $noteframe->Label(-text => "current value")
    ->pack(-side => 'left', -fill => "x");
$note4 = $noteframe->Label(-textvariable => \$cnbk, -bg => "#ffccff")
    ->pack(-side => 'left', -fill => "x");

$mapframe = $mw->Frame(-relief => 'sunken', -bd => 2, -width => $cwid)
    ->pack(-side => 'top');

$map  = $mapframe->Canvas(-bg => "white", -width => $cwid, -height => $cht)
    ->pack(-side => 'bottom');

if($nbk%$nfs==0) {
    $tmp_ht = int($nbk/$nfs);
}
else {
    $tmp_ht = int($nbk/$nfs) + 1;
}

if($cfgname eq '') {
	$rest = $nbk;
	for($i=0;$i<$nfs;$i++) {
		if($rest>$tmp_ht) {
			$bk[$i] = $tmp_ht;
			$rest -= $tmp_ht;
		}
		else {
			$bk[$i] = $rest;
			$rest = 0;
		}
#		print "$i $bk[$i] $rest\n";
	}
}
else {
	open(CF, "<$cfgname") || die "cannot open configuration file '$cfgname'";
	while(<CF>) {
		if(/^#/) {
			next;
		}
		@f = split;
		if($f[0] =~ /fsmap/i ) {
#			print STDERR "fs# $f[1], bk range $f[2]\n";
			$c = 0;
			foreach $r (split /,/, $f[2]) {
				if($r =~ /^(\d+)-(\d+)$/) {
					$s = $1;
					$e = $2;
#					print STDERR "\trange $r ; $s to $e\n";
					$c += $2-$1+1;
				}
				elsif($r =~ /^(\d+)$/) {
					$s = $1;
					$e = $1;
#					print STDERR "\tsimple $r ; $s\n";
					$c += 1;
				}
				else {
					$s = -1;
				}
			}
			$bk[$f[1]] = $c;
#			print STDERR "\t\t$c bk\n";
		}
	}
	close(CF);

}

	for($i=0;$i<$nfs;$i++) {
		$ibk[$i] = $bk[$i];
	}

	$cnbk = 0;
	for($i=0;$i<$nfs;$i++) {
		$cnbk += $bk[$i];

		$tmp_x = $zx + $i*$fwid;
		$tmp_y = $zy + $tht;

		$tag = $i ."lb";
		$lawg[$i] = $map->createText($tmp_x+$fwid/2, $tmp_y+$tht/2,
			-text => "$bk[$i]", -tags => $tag);
		$rlawg[$lawg[$i]] = $i;

		$tag = sprintf("fs#%d", $i);
		$fswg[$i] = $map->createRectangle($tmp_x, $tmp_y,
				$tmp_x+$fwid-1, $tmp_y-$bk[$i]*$fht,
				-fill => "#ffccff", -tags => $tag );
		$rfswg[$fswg[$i]] = $i;
		
#		print "$i / $fswg[$i] / $tag\n";
	}

sub redraw {
	$cnbk = 0;
	for($i=0;$i<$nfs;$i++) {
		$cnbk += $bk[$i];

		$tmp_x = $zx + $i*$fwid;
		$tmp_y = $zy + $tht;

		$map->itemconfigure($lawg[$i], -text => "$bk[$i]");
		$map->coords($fswg[$i], $tmp_x, $tmp_y, $tmp_x+$fwid-1, $tmp_y-$bk[$i]*$fht);
	}
}


# @list = $map->find("all");
# print "tags =",join(":",@list),"\n";

#$map->Tk::bind("all", "<Button-1>", [ \&mark,Ev('x'),Ev('y')] );
$map->Tk::bind("all", "<Button-1>", [ \&up,Ev('x'),Ev('y')] );
$map->Tk::bind("all", "<Button-2>", [ \&beflat,Ev('x'),Ev('y')] );
$map->Tk::bind("all", "<Button-3>", [ \&down,Ev('x'),Ev('y')] );


MainLoop;

sub mark {
    my($w, $x, $y) = @_;
    my($id);
    my($f);

#   print "$w / $x / $y\n";

    if($w!=$map) {
        return;
    }

    $id = $map->find("closest", $x, $y);
    $f = $rfswg[$id];
#    print "  target $id , $f\n";

    &redraw;
}

sub up {
    my($w, $x, $y) = @_;
    my($id);
    my($f);

#   print "$w / $x / $y\n";

    if($w!=$map) {
        return;
    }

    $id = $map->find("closest", $x, $y);
    $f = $rfswg[$id];
#    print "  target $id , $f\n";

    if($bk[$f]<$nbk) {
        $bk[$f]++;

        &redraw;
    }
}

sub down {
    my($w, $x, $y) = @_;
    my($id);
    my($f);

#   print "$w / $x / $y\n";

    if($w!=$map) {
        return;
    }

    $id = $map->find("closest", $x, $y);
    $f = $rfswg[$id];
#    print "  target $id , $f\n";

    if($bk[$f]>0) {
        $bk[$f]--;

        &redraw;
    }
}

sub beflat {
    my($w, $x, $y) = @_;
    my($id);
    my($f);
    my($i);

#   print "$w / $x / $y\n";

    if($w!=$map) {
        return;
    }

    $id = $map->find("closest", $x, $y);
    $f = $rfswg[$id];
#    print "  target $id , $f\n";

    for($i=0;$i<$nfs;$i++) {
        if($i!=$f) {
            $bk[$i] = $bk[$f];
        }
    }

    &redraw;
}


sub fsprint {
    my($i, $s);
    
    print "#\n# generated by mkfsmap.pl; nfs $nfs, nbk $nbk, cnbk $cnbk\n#\n";
    if($cnbk!=$nbk) {
    print "#\n# NOTICE: this configuration is not correct\n#\n";
    }

    $s = 0;
    for($i=0;$i<$nfs;$i++) {
#        printf "fsmap %2d %3d %3d\n", $i, $s, $s+$bk[$i]-1;
        printf "fsmap %2d %d-%d\n", $i, $s, $s+$bk[$i]-1;
        $s += $bk[$i];
    }
}

sub reset {
	my($i);

	for($i=0;$i<$nfs;$i++) {
		$bk[$i] = $ibk[$i];
	}

	&redraw;
}
