/*
 * KOTETU:
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>

#include <sys/types.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif

#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

#include "range.h"
#include "msg.h"
/*
#include "trace.h"
*/
#include "mem.h"
#include "mlist.h"

#include "proxy_global.h"

#include "args.h"
#include "http.h"
#include "sim_stat.h"
#include "sess.h"
#include "ocache.h"

/*
 * externals
 */

/* automatic */
extern                       char config_filename[];
extern                      char* log_filename;
extern                        int trace_flag;
extern                        int monitor_flag;
extern                        int prefetch_flag;
extern                        int fdcaching_flag;
extern                        int bitmapping_flag;
extern                        int service_portno;
extern                      char* localdomain_padding;
extern                      char* report_hook_hostname;
extern                        int require_nsess_total;
extern                        int require_n_handler;
extern                        int require_nsess_handler;
extern                      char* cache_dir_prefix;
extern                      char* cache_index_prefix;
extern                        int require_radix_h2;
extern                        int require_radix_h3;
extern                        int require_disk_sweep_interval;
extern                        int require_nfrag_obj;
extern                        int require_nfrag_nwrite;
extern                        int require_nfrag_nread;
extern                        int require_nfrag_dwrite;
extern                        int require_nfrag_dread;
extern                        int require_nindex_dread;
extern                        int require_nindex_bdread;
extern                        int require_frag_hhwater;
extern                        int require_frag_hwater;
extern                        int require_frag_lwater;
extern                        int require_entry_hhwater;
extern                        int require_entry_hwater;
extern                        int require_entry_lwater;
extern                        int require_n_bsweeper;
extern                        int require_n_msweeper;
extern                        int require_n_fswapper;
extern                      off_t require_total_memory_size;
extern                        int require_nbucket;
extern                        int require_nentry_per_bucket;
extern                        int require_level_bucket_skiplist;
extern                        int require_frag_size;
extern                        int num_prefetch_referedpages;
extern                        int num_prefetch_includedobjs;
extern                        int require_thru_factor;

/* static */
extern pid_t    wcol_pid;
extern char     *progname;

/* end of externals */

char config_file[STRING_SIZE];
char dummy_string[STRING_SIZE];


/*
 * SetGlobalDefaults -
 */
void
SetGlobalDefaults()
{
    strcpy(config_filename, DEFAULT_CONFIGFILE);
    log_filename = strdup(DEFAULT_LOGFILE);
    trace_flag = 1;
    monitor_flag = 1;
    prefetch_flag = 1;
    fdcaching_flag = 1;
    bitmapping_flag = 1;
    service_portno = (int) HTTP_PROXYPORTNO;
    localdomain_padding = strdup(".kotetu.cache");
    report_hook_hostname = strdup("www.self-report.cache");
    require_nsess_total = (int) 2048;
    require_n_handler = (int) 4;
    require_nsess_handler = (int) 32;
    cache_dir_prefix = strdup(DEFAULT_CACHEDIR);
    cache_index_prefix = strdup("");
    require_radix_h2 = (int) 17;
    require_radix_h3 = (int) 19;
    require_disk_sweep_interval = (int) 1000;
    require_nfrag_obj = (int) 256;
    require_nfrag_nwrite = (int) 4;
    require_nfrag_nread = (int) 4;
    require_nfrag_dwrite = (int) 4;
    require_nfrag_dread = (int) 4;
    require_nindex_dread = (int) 16;
    require_nindex_bdread = (int) 32;
    require_frag_hhwater = (int) 90;
    require_frag_hwater = (int) 80;
    require_frag_lwater = (int) 60;
    require_entry_hhwater = (int) 90;
    require_entry_hwater = (int) 80;
    require_entry_lwater = (int) 60;
    require_n_bsweeper = (int) 1;
    require_n_msweeper = (int) 1;
    require_n_fswapper = (int) 8;
    require_total_memory_size = (off_t) 32*(1048576);
    require_nbucket = (int) 53;
    require_nentry_per_bucket = (int) 64;
    require_level_bucket_skiplist = (int) 8;
    require_frag_size = (int) 4096;
    num_prefetch_referedpages = (int) 8;
    num_prefetch_includedobjs = (int) 16;
    require_thru_factor = (int) -1;
}
/* end of SetGlobalDefaults() */





#define MAX_AV 10

/*
 * ReadConfigFile -
 */
void ReadConfigFile(char *filename)
{
    arg_t  *args, *pargs;
    int     num;
    int     tint;
    char   *av[MAX_AV];
    int     i, j;


    Trace("ReadConfigFile: filename='%s'\n", filename);
    if((num=ReadArgsFromFile(filename, &args))<=0) {
        Trace("ReadConfigFile: no arguments\n");
        return;
    }

    pargs = args;
    while(pargs) {
        /* project argument value to temporary pointor */
        for(i=0;pargs->vals[i] && i<MAX_AV;i++) {
            av[i] = pargs->vals[i];
        }
        for(j=i;j<MAX_AV;j++) {
            av[j] = NULL;
        }

        if(!pargs->name) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "ConfigFile")) {
            if(pargs->vals[0]) {
                strcpy(config_filename,av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LogFile")) {
            if(pargs->vals[0]) {
                log_filename = strdup(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'T', "Trace")) {
            trace_flag = 1;
            goto nextone;
        }
        else
        if((pargs->name[0]=='N' || pargs->name[0]=='n')
            && (pargs->name[1]=='O' || pargs->name[1]=='o')
            && strcasecmp(&pargs->name[2], "Trace")==0) {
            trace_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "Monitor")) {
            monitor_flag = 1;
            goto nextone;
        }
        else
        if((pargs->name[0]=='N' || pargs->name[0]=='n')
            && (pargs->name[1]=='O' || pargs->name[1]=='o')
            && strcasecmp(&pargs->name[2], "Monitor")==0) {
            monitor_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "Prefetch")) {
            prefetch_flag = 1;
            goto nextone;
        }
        else
        if((pargs->name[0]=='N' || pargs->name[0]=='n')
            && (pargs->name[1]=='O' || pargs->name[1]=='o')
            && strcasecmp(&pargs->name[2], "Prefetch")==0) {
            prefetch_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FDcaching")) {
            fdcaching_flag = 1;
            goto nextone;
        }
        else
        if((pargs->name[0]=='N' || pargs->name[0]=='n')
            && (pargs->name[1]=='O' || pargs->name[1]=='o')
            && strcasecmp(&pargs->name[2], "FDcaching")==0) {
            fdcaching_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'B', "Bitmapping")) {
            bitmapping_flag = 1;
            goto nextone;
        }
        else
        if((pargs->name[0]=='N' || pargs->name[0]=='n')
            && (pargs->name[1]=='O' || pargs->name[1]=='o')
            && strcasecmp(&pargs->name[2], "Bitmapping")==0) {
            bitmapping_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "ServicePort")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    service_portno = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LocalDomain")) {
            if(pargs->vals[0]) {
                localdomain_padding = strdup(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "ReportHook")) {
            if(pargs->vals[0]) {
                report_hook_hostname = strdup(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumSession")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nsess_total = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumHandler")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_n_handler = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NSessHandler")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nsess_handler = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheDir")) {
            if(pargs->vals[0]) {
                cache_dir_prefix = strdup(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheIndexDir")) {
            if(pargs->vals[0]) {
                cache_index_prefix = strdup(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RadixH2")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_radix_h2 = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RadixH3")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_radix_h3 = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'D', "DiskSweepInterval")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_disk_sweep_interval = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "Fsmap")) {
            if(pargs->vals[0]) {
                if(parse_fsmap(av[0], av[1])==0) {
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NFragObject")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nfrag_obj = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NFragNetWrite")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nfrag_nwrite = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NFragNetRead")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nfrag_nread = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NFragDiskWrite")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nfrag_dwrite = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NFragDiskRead")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nfrag_dread = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NIndexDiskRead")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nindex_dread = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NIndexBurstDiskRead")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nindex_bdread = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FragHHwater")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_frag_hhwater = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FragHwater")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_frag_hwater = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FragLwater")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_frag_lwater = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'E', "EntryHHwater")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_entry_hhwater = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'E', "EntryHwater")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_entry_hwater = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'E', "EntryLwater")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_entry_lwater = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumDiskSweeper")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_n_bsweeper = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumMemorySweeper")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_n_msweeper = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumSwapper")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_n_fswapper = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'T', "TotalMemorySize")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_total_memory_size = (off_t) tint * 1048576;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumBucket")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nbucket = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NEntryBucket")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_nentry_per_bucket = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LevelBucketSkipList")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_level_bucket_skiplist = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FragSize")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_frag_size = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumPrefetchReferedPages")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    num_prefetch_referedpages = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumPrefetchIncludedObjects")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    num_prefetch_includedobjs = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'A', "AcceptAddress")) {
            if(pargs->vals[0]) {
                if(AppendAcceptEntry(av[0], av[1])==0) {
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "Parent")) {
            if(pargs->vals[0]) {
                if(AppendParent(av[0])==0) {
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'T', "ThruFactor")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_thru_factor = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RadixH1")) {
            /* nothing */
            goto nextone;
        }
        else {
            Error("Unknow argument '%s'", pargs->name);
            goto nextone;
        }
fail:
        Error("Argument specification ('%s') is wrong", pargs->name);
nextone:
        pargs = pargs->next;
    }

    FreeArgs(args);
}



/*
 * FixGlobals - Fix global vars.
 */
void
FixGlobals()
{

    if(service_portno<1) {
        Error("FixGlobals: 'ServicePort' is too small (min %d)",1);
        service_portno = (int) 1;
    }
    if(service_portno>32000) {
        Error("FixGlobals: 'ServicePort' is too large (max %d)",32000);
        service_portno = (int) 32000;
    }
    if(require_nsess_total<128) {
        Error("FixGlobals: 'NumSession' is too small (min %d)",128);
        require_nsess_total = (int) 128;
    }
    if(require_nsess_total>NSESS_TOTAL_MAX) {
        Error("FixGlobals: 'NumSession' is too large (max %d)",NSESS_TOTAL_MAX);
        require_nsess_total = (int) NSESS_TOTAL_MAX;
    }
    if(require_n_handler<1) {
        Error("FixGlobals: 'NumHandler' is too small (min %d)",1);
        require_n_handler = (int) 1;
    }
    if(require_n_handler>N_HANDLER_MAX) {
        Error("FixGlobals: 'NumHandler' is too large (max %d)",N_HANDLER_MAX);
        require_n_handler = (int) N_HANDLER_MAX;
    }
    if(require_nsess_handler<1) {
        Error("FixGlobals: 'NSessHandler' is too small (min %d)",1);
        require_nsess_handler = (int) 1;
    }
    if(require_nsess_handler>NSESS_HANDLER_MAX) {
        Error("FixGlobals: 'NSessHandler' is too large (max %d)",NSESS_HANDLER_MAX);
        require_nsess_handler = (int) NSESS_HANDLER_MAX;
    }
    if(require_radix_h2<3) {
        Error("FixGlobals: 'RadixH2' is too small (min %d)",3);
        require_radix_h2 = (int) 3;
    }
    if(require_radix_h2>RADIX_H2_MAX) {
        Error("FixGlobals: 'RadixH2' is too large (max %d)",RADIX_H2_MAX);
        require_radix_h2 = (int) RADIX_H2_MAX;
    }
    if(require_radix_h3<3) {
        Error("FixGlobals: 'RadixH3' is too small (min %d)",3);
        require_radix_h3 = (int) 3;
    }
    if(require_radix_h3>RADIX_H3_MAX) {
        Error("FixGlobals: 'RadixH3' is too large (max %d)",RADIX_H3_MAX);
        require_radix_h3 = (int) RADIX_H3_MAX;
    }
    if(require_disk_sweep_interval<100) {
        Error("FixGlobals: 'DiskSweepInterval' is too small (min %d)",100);
        require_disk_sweep_interval = (int) 100;
    }
    if(require_disk_sweep_interval>3600000) {
        Error("FixGlobals: 'DiskSweepInterval' is too large (max %d)",3600000);
        require_disk_sweep_interval = (int) 3600000;
    }
    if(require_nfrag_obj<1) {
        Error("FixGlobals: 'NFragObject' is too small (min %d)",1);
        require_nfrag_obj = (int) 1;
    }
    if(require_nfrag_obj>HARD_NFRAG_OBJ) {
        Error("FixGlobals: 'NFragObject' is too large (max %d)",HARD_NFRAG_OBJ);
        require_nfrag_obj = (int) HARD_NFRAG_OBJ;
    }
    if(require_nfrag_nwrite<1) {
        Error("FixGlobals: 'NFragNetWrite' is too small (min %d)",1);
        require_nfrag_nwrite = (int) 1;
    }
    if(require_nfrag_nwrite>HARD_NFRAG_NWRITE) {
        Error("FixGlobals: 'NFragNetWrite' is too large (max %d)",HARD_NFRAG_NWRITE);
        require_nfrag_nwrite = (int) HARD_NFRAG_NWRITE;
    }
    if(require_nfrag_nread<1) {
        Error("FixGlobals: 'NFragNetRead' is too small (min %d)",1);
        require_nfrag_nread = (int) 1;
    }
    if(require_nfrag_nread>HARD_NFRAG_NREAD) {
        Error("FixGlobals: 'NFragNetRead' is too large (max %d)",HARD_NFRAG_NREAD);
        require_nfrag_nread = (int) HARD_NFRAG_NREAD;
    }
    if(require_nfrag_dwrite<1) {
        Error("FixGlobals: 'NFragDiskWrite' is too small (min %d)",1);
        require_nfrag_dwrite = (int) 1;
    }
    if(require_nfrag_dwrite>HARD_NFRAG_DWRITE) {
        Error("FixGlobals: 'NFragDiskWrite' is too large (max %d)",HARD_NFRAG_DWRITE);
        require_nfrag_dwrite = (int) HARD_NFRAG_DWRITE;
    }
    if(require_nfrag_dread<1) {
        Error("FixGlobals: 'NFragDiskRead' is too small (min %d)",1);
        require_nfrag_dread = (int) 1;
    }
    if(require_nfrag_dread>HARD_NFRAG_DREAD) {
        Error("FixGlobals: 'NFragDiskRead' is too large (max %d)",HARD_NFRAG_DREAD);
        require_nfrag_dread = (int) HARD_NFRAG_DREAD;
    }
    if(require_nindex_dread<1) {
        Error("FixGlobals: 'NIndexDiskRead' is too small (min %d)",1);
        require_nindex_dread = (int) 1;
    }
    if(require_nindex_dread>HARD_NINDEX_DREAD) {
        Error("FixGlobals: 'NIndexDiskRead' is too large (max %d)",HARD_NINDEX_DREAD);
        require_nindex_dread = (int) HARD_NINDEX_DREAD;
    }
    if(require_nindex_bdread<1) {
        Error("FixGlobals: 'NIndexBurstDiskRead' is too small (min %d)",1);
        require_nindex_bdread = (int) 1;
    }
    if(require_nindex_bdread>HARD_NINDEX_BDREAD) {
        Error("FixGlobals: 'NIndexBurstDiskRead' is too large (max %d)",HARD_NINDEX_BDREAD);
        require_nindex_bdread = (int) HARD_NINDEX_BDREAD;
    }
    if(require_frag_hhwater<1) {
        Error("FixGlobals: 'FragHHwater' is too small (min %d)",1);
        require_frag_hhwater = (int) 1;
    }
    if(require_frag_hhwater>100) {
        Error("FixGlobals: 'FragHHwater' is too large (max %d)",100);
        require_frag_hhwater = (int) 100;
    }
    if(require_frag_hwater<1) {
        Error("FixGlobals: 'FragHwater' is too small (min %d)",1);
        require_frag_hwater = (int) 1;
    }
    if(require_frag_hwater>100) {
        Error("FixGlobals: 'FragHwater' is too large (max %d)",100);
        require_frag_hwater = (int) 100;
    }
    if(require_frag_lwater<1) {
        Error("FixGlobals: 'FragLwater' is too small (min %d)",1);
        require_frag_lwater = (int) 1;
    }
    if(require_frag_lwater>100) {
        Error("FixGlobals: 'FragLwater' is too large (max %d)",100);
        require_frag_lwater = (int) 100;
    }
    if(require_entry_hhwater<1) {
        Error("FixGlobals: 'EntryHHwater' is too small (min %d)",1);
        require_entry_hhwater = (int) 1;
    }
    if(require_entry_hhwater>100) {
        Error("FixGlobals: 'EntryHHwater' is too large (max %d)",100);
        require_entry_hhwater = (int) 100;
    }
    if(require_entry_hwater<1) {
        Error("FixGlobals: 'EntryHwater' is too small (min %d)",1);
        require_entry_hwater = (int) 1;
    }
    if(require_entry_hwater>100) {
        Error("FixGlobals: 'EntryHwater' is too large (max %d)",100);
        require_entry_hwater = (int) 100;
    }
    if(require_entry_lwater<1) {
        Error("FixGlobals: 'EntryLwater' is too small (min %d)",1);
        require_entry_lwater = (int) 1;
    }
    if(require_entry_lwater>100) {
        Error("FixGlobals: 'EntryLwater' is too large (max %d)",100);
        require_entry_lwater = (int) 100;
    }
    if(require_n_bsweeper<1) {
        Error("FixGlobals: 'NumDiskSweeper' is too small (min %d)",1);
        require_n_bsweeper = (int) 1;
    }
    if(require_n_bsweeper>32) {
        Error("FixGlobals: 'NumDiskSweeper' is too large (max %d)",32);
        require_n_bsweeper = (int) 32;
    }
    if(require_n_msweeper<1) {
        Error("FixGlobals: 'NumMemorySweeper' is too small (min %d)",1);
        require_n_msweeper = (int) 1;
    }
    if(require_n_msweeper>32) {
        Error("FixGlobals: 'NumMemorySweeper' is too large (max %d)",32);
        require_n_msweeper = (int) 32;
    }
    if(require_n_fswapper<1) {
        Error("FixGlobals: 'NumSwapper' is too small (min %d)",1);
        require_n_fswapper = (int) 1;
    }
    if(require_n_fswapper>N_SWAPPER_MAX) {
        Error("FixGlobals: 'NumSwapper' is too large (max %d)",N_SWAPPER_MAX);
        require_n_fswapper = (int) N_SWAPPER_MAX;
    }
    if(require_nbucket<1) {
        Error("FixGlobals: 'NumBucket' is too small (min %d)",1);
        require_nbucket = (int) 1;
    }
    if(require_nbucket>N_BUCKET_MAX) {
        Error("FixGlobals: 'NumBucket' is too large (max %d)",N_BUCKET_MAX);
        require_nbucket = (int) N_BUCKET_MAX;
    }
    if(require_nentry_per_bucket<4) {
        Error("FixGlobals: 'NEntryBucket' is too small (min %d)",4);
        require_nentry_per_bucket = (int) 4;
    }
    if(require_nentry_per_bucket>1024) {
        Error("FixGlobals: 'NEntryBucket' is too large (max %d)",1024);
        require_nentry_per_bucket = (int) 1024;
    }
    if(require_level_bucket_skiplist<2) {
        Error("FixGlobals: 'LevelBucketSkipList' is too small (min %d)",2);
        require_level_bucket_skiplist = (int) 2;
    }
    if(require_level_bucket_skiplist>64) {
        Error("FixGlobals: 'LevelBucketSkipList' is too large (max %d)",64);
        require_level_bucket_skiplist = (int) 64;
    }
    if(require_frag_size<128) {
        Error("FixGlobals: 'FragSize' is too small (min %d)",128);
        require_frag_size = (int) 128;
    }
    if(require_frag_size>8192) {
        Error("FixGlobals: 'FragSize' is too large (max %d)",8192);
        require_frag_size = (int) 8192;
    }
    if(num_prefetch_referedpages<0) {
        Error("FixGlobals: 'NumPrefetchReferedPages' is too small (min %d)",0);
        num_prefetch_referedpages = (int) 0;
    }
    if(num_prefetch_referedpages>64) {
        Error("FixGlobals: 'NumPrefetchReferedPages' is too large (max %d)",64);
        num_prefetch_referedpages = (int) 64;
    }
    if(num_prefetch_includedobjs<0) {
        Error("FixGlobals: 'NumPrefetchIncludedObjects' is too small (min %d)",0);
        num_prefetch_includedobjs = (int) 0;
    }
    if(num_prefetch_includedobjs>128) {
        Error("FixGlobals: 'NumPrefetchIncludedObjects' is too large (max %d)",128);
        num_prefetch_includedobjs = (int) 128;
    }
    if(require_thru_factor<-1) {
        Error("FixGlobals: 'ThruFactor' is too small (min %d)",-1);
        require_thru_factor = (int) -1;
    }
    if(require_thru_factor>100) {
        Error("FixGlobals: 'ThruFactor' is too large (max %d)",100);
        require_thru_factor = (int) 100;
    }
}




void
ShowGlobals(FILE *s)
{
    fprintf(s,
"======================================================================\n\
Flag Name                            CurrentValue\n\
----------------------------------------------------------------------\n");

    fprintf(s, "  f  ConfigFile                      '%s'\n",
        config_filename);
    fprintf(s, "  l  LogFile                         '%s'\n",
        log_filename);
    fprintf(s, "  t  Trace                            %s\n",
        trace_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  Monitor                          %s\n",
        monitor_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  P  Prefetch                         %s\n",
        prefetch_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  FDcaching                        %s\n",
        fdcaching_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  Bitmapping                       %s\n",
        bitmapping_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  p  ServicePort                     %8d \n",
        service_portno);
    fprintf(s, "  -  LocalDomain                     '%s'\n",
        localdomain_padding);
    fprintf(s, "  R  ReportHook                      '%s'\n",
        report_hook_hostname);
    fprintf(s, "  n  NumSession                      %8d \n",
        require_nsess_total);
    fprintf(s, "  H  NumHandler                      %8d \n",
        require_n_handler);
    fprintf(s, "  C  NSessHandler                    %8d \n",
        require_nsess_handler);
    fprintf(s, "  c  CacheDir                        '%s'\n",
        cache_dir_prefix);
    fprintf(s, "  -  CacheIndexDir                   '%s'\n",
        cache_index_prefix);
    fprintf(s, "  -  RadixH2                         %8d \n",
        require_radix_h2);
    fprintf(s, "  -  RadixH3                         %8d \n",
        require_radix_h3);
    fprintf(s, "  -  DiskSweepInterval               %8d \n",
        require_disk_sweep_interval);
    fprintf(s, "  -  NFragObject                     %8d \n",
        require_nfrag_obj);
    fprintf(s, "  -  NFragNetWrite                   %8d \n",
        require_nfrag_nwrite);
    fprintf(s, "  -  NFragNetRead                    %8d \n",
        require_nfrag_nread);
    fprintf(s, "  -  NFragDiskWrite                  %8d \n",
        require_nfrag_dwrite);
    fprintf(s, "  -  NFragDiskRead                   %8d \n",
        require_nfrag_dread);
    fprintf(s, "  -  NIndexDiskRead                  %8d \n",
        require_nindex_dread);
    fprintf(s, "  -  NIndexBurstDiskRead             %8d \n",
        require_nindex_bdread);
    fprintf(s, "  -  FragHHwater                     %8d \n",
        require_frag_hhwater);
    fprintf(s, "  -  FragHwater                      %8d \n",
        require_frag_hwater);
    fprintf(s, "  -  FragLwater                      %8d \n",
        require_frag_lwater);
    fprintf(s, "  -  EntryHHwater                    %8d \n",
        require_entry_hhwater);
    fprintf(s, "  -  EntryHwater                     %8d \n",
        require_entry_hwater);
    fprintf(s, "  -  EntryLwater                     %8d \n",
        require_entry_lwater);
    fprintf(s, "  D  NumDiskSweeper                  %8d \n",
        require_n_bsweeper);
    fprintf(s, "  M  NumMemorySweeper                %8d \n",
        require_n_msweeper);
    fprintf(s, "  W  NumSwapper                      %8d \n",
        require_n_fswapper);
    fprintf(s, "  U  TotalMemorySize                 %8d M (x 1048576)\n",
        require_total_memory_size/(1048576));
    fprintf(s, "  -  NumBucket                       %8d \n",
        require_nbucket);
    fprintf(s, "  -  NEntryBucket                    %8d \n",
        require_nentry_per_bucket);
    fprintf(s, "  -  LevelBucketSkipList             %8d \n",
        require_level_bucket_skiplist);
    fprintf(s, "  -  FragSize                        %8d \n",
        require_frag_size);
    fprintf(s, "  r  NumPrefetchReferedPages         %8d \n",
        num_prefetch_referedpages);
    fprintf(s, "  i  NumPrefetchIncludedObjects      %8d \n",
        num_prefetch_includedobjs);
    fprintf(s, "  T  ThruFactor                      %8d \n",
        require_thru_factor);
}

void
ShowGlobals_fd(int ofd)
{
    fdprintf(ofd,
"======================================================================\n\
Flag Name                        CurrentValue\n\
----------------------------------------------------------------------\n");

    fdprintf(ofd, "  f  ConfigFile                  '%s'\n",
        config_filename);
    fdprintf(ofd, "  l  LogFile                     '%s'\n",
        log_filename);
    fdprintf(ofd, "  t  Trace                        %s\n",
        trace_flag==1 ? "*ON*" : "*OFF*" );
    fdprintf(ofd, "  -  Monitor                      %s\n",
        monitor_flag==1 ? "*ON*" : "*OFF*" );
    fdprintf(ofd, "  P  Prefetch                     %s\n",
        prefetch_flag==1 ? "*ON*" : "*OFF*" );
    fdprintf(ofd, "  -  FDcaching                    %s\n",
        fdcaching_flag==1 ? "*ON*" : "*OFF*" );
    fdprintf(ofd, "  -  Bitmapping                   %s\n",
        bitmapping_flag==1 ? "*ON*" : "*OFF*" );
    fdprintf(ofd, "  p  ServicePort                 %8d \n",
        service_portno);
    fdprintf(ofd, "  -  LocalDomain                 '%s'\n",
        localdomain_padding);
    fdprintf(ofd, "  R  ReportHook                  '%s'\n",
        report_hook_hostname);
    fdprintf(ofd, "  n  NumSession                  %8d \n",
        require_nsess_total);
    fdprintf(ofd, "  H  NumHandler                  %8d \n",
        require_n_handler);
    fdprintf(ofd, "  C  NSessHandler                %8d \n",
        require_nsess_handler);
    fdprintf(ofd, "  c  CacheDir                    '%s'\n",
        cache_dir_prefix);
    fdprintf(ofd, "  -  CacheIndexDir               '%s'\n",
        cache_index_prefix);
    fdprintf(ofd, "  -  RadixH2                     %8d \n",
        require_radix_h2);
    fdprintf(ofd, "  -  RadixH3                     %8d \n",
        require_radix_h3);
    fdprintf(ofd, "  -  DiskSweepInterval           %8d \n",
        require_disk_sweep_interval);
    fdprintf(ofd, "  -  NFragObject                 %8d \n",
        require_nfrag_obj);
    fdprintf(ofd, "  -  NFragNetWrite               %8d \n",
        require_nfrag_nwrite);
    fdprintf(ofd, "  -  NFragNetRead                %8d \n",
        require_nfrag_nread);
    fdprintf(ofd, "  -  NFragDiskWrite              %8d \n",
        require_nfrag_dwrite);
    fdprintf(ofd, "  -  NFragDiskRead               %8d \n",
        require_nfrag_dread);
    fdprintf(ofd, "  -  NIndexDiskRead              %8d \n",
        require_nindex_dread);
    fdprintf(ofd, "  -  NIndexBurstDiskRead         %8d \n",
        require_nindex_bdread);
    fdprintf(ofd, "  -  FragHHwater                 %8d \n",
        require_frag_hhwater);
    fdprintf(ofd, "  -  FragHwater                  %8d \n",
        require_frag_hwater);
    fdprintf(ofd, "  -  FragLwater                  %8d \n",
        require_frag_lwater);
    fdprintf(ofd, "  -  EntryHHwater                %8d \n",
        require_entry_hhwater);
    fdprintf(ofd, "  -  EntryHwater                 %8d \n",
        require_entry_hwater);
    fdprintf(ofd, "  -  EntryLwater                 %8d \n",
        require_entry_lwater);
    fdprintf(ofd, "  D  NumDiskSweeper              %8d \n",
        require_n_bsweeper);
    fdprintf(ofd, "  M  NumMemorySweeper            %8d \n",
        require_n_msweeper);
    fdprintf(ofd, "  W  NumSwapper                  %8d \n",
        require_n_fswapper);
    fdprintf(ofd, "  U  TotalMemorySize             %8d M (x 1048576)\n",
        require_total_memory_size/(1048576));
    fdprintf(ofd, "  -  NumBucket                   %8d \n",
        require_nbucket);
    fdprintf(ofd, "  -  NEntryBucket                %8d \n",
        require_nentry_per_bucket);
    fdprintf(ofd, "  -  LevelBucketSkipList         %8d \n",
        require_level_bucket_skiplist);
    fdprintf(ofd, "  -  FragSize                    %8d \n",
        require_frag_size);
    fdprintf(ofd, "  r  NumPrefetchReferedPages     %8d \n",
        num_prefetch_referedpages);
    fdprintf(ofd, "  i  NumPrefetchIncludedObjects  %8d \n",
        num_prefetch_includedobjs);
    fdprintf(ofd, "  T  ThruFactor                  %8d \n",
        require_thru_factor);
}

print_configration_FORM(int ofd)
{
    fdprintf(ofd,
"<form name=\"chk\" method=\"POST\" action=\"/config\">\n");

    fdprintf(ofd,
"<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\">\n\
<tr bgcolor=\"#ffffcc\">\n\
    <th rowspan=\"2\"> name\n\
    <th colspan=\"2\"> current\n\
    <th rowspan=\"2\"> specified\n\
    <th rowspan=\"2\"> descriptions\n\
<tr bgcolor=\"#ffffcc\">\n\
    <th> value\n\
    <th> unit\n\
\n\
");

    fdprintf(ofd, "<tr><td>ConfigFile  <td colspan=\"2\"> %s <td><input name=\"ConfigFile\" value=\"%s\"> <td> path\n",
        config_filename, config_filename);
    fdprintf(ofd, "<tr><td>LogFile  <td colspan=\"2\"> %s <td><input name=\"LogFile\" value=\"%s\"> <td> path\n",
        log_filename, log_filename);
    fdprintf(ofd, "<tr><td>Trace  <td colspan=\"2\"> %s <td><input name=\"Trace\" value=\"%s\"> <td> on/off\n",
        trace_flag==1 ? "on" : "off",
        trace_flag==1 ? "on" : "off" );
    fdprintf(ofd, "<tr><td>Monitor  <td colspan=\"2\"> %s <td><input name=\"Monitor\" value=\"%s\"> <td> on/off\n",
        monitor_flag==1 ? "on" : "off",
        monitor_flag==1 ? "on" : "off" );
    fdprintf(ofd, "<tr><td>Prefetch  <td colspan=\"2\"> %s <td><input name=\"Prefetch\" value=\"%s\"> <td> on/off\n",
        prefetch_flag==1 ? "on" : "off",
        prefetch_flag==1 ? "on" : "off" );
    fdprintf(ofd, "<tr><td>FDcaching  <td colspan=\"2\"> %s <td><input name=\"FDcaching\" value=\"%s\"> <td> on/off\n",
        fdcaching_flag==1 ? "on" : "off",
        fdcaching_flag==1 ? "on" : "off" );
    fdprintf(ofd, "<tr><td>Bitmapping  <td colspan=\"2\"> %s <td><input name=\"Bitmapping\" value=\"%s\"> <td> on/off\n",
        bitmapping_flag==1 ? "on" : "off",
        bitmapping_flag==1 ? "on" : "off" );
    fdprintf(ofd, "<tr><td>ServicePort  <td align=\"right\"> %8d <td>  <td><input name=\"ServicePort\" value=\"%d\"> <td> max %d min %d\n",
        service_portno, service_portno, 32000, 1);
    fdprintf(ofd, "<tr><td>LocalDomain  <td colspan=\"2\"> %s <td><input name=\"LocalDomain\" value=\"%s\"> <td> path\n",
        localdomain_padding, localdomain_padding);
    fdprintf(ofd, "<tr><td>ReportHook  <td colspan=\"2\"> %s <td><input name=\"ReportHook\" value=\"%s\"> <td> path\n",
        report_hook_hostname, report_hook_hostname);
    fdprintf(ofd, "<tr><td>NumSession  <td align=\"right\"> %8d <td>  <td><input name=\"NumSession\" value=\"%d\"> <td> max %d min %d\n",
        require_nsess_total, require_nsess_total, NSESS_TOTAL_MAX, 128);
    fdprintf(ofd, "<tr><td>NumHandler  <td align=\"right\"> %8d <td>  <td><input name=\"NumHandler\" value=\"%d\"> <td> max %d min %d\n",
        require_n_handler, require_n_handler, N_HANDLER_MAX, 1);
    fdprintf(ofd, "<tr><td>NSessHandler  <td align=\"right\"> %8d <td>  <td><input name=\"NSessHandler\" value=\"%d\"> <td> max %d min %d\n",
        require_nsess_handler, require_nsess_handler, NSESS_HANDLER_MAX, 1);
    fdprintf(ofd, "<tr><td>CacheDir  <td colspan=\"2\"> %s <td><input name=\"CacheDir\" value=\"%s\"> <td> path\n",
        cache_dir_prefix, cache_dir_prefix);
    fdprintf(ofd, "<tr><td>CacheIndexDir  <td colspan=\"2\"> %s <td><input name=\"CacheIndexDir\" value=\"%s\"> <td> path\n",
        cache_index_prefix, cache_index_prefix);
    fdprintf(ofd, "<tr><td>RadixH2  <td align=\"right\"> %8d <td>  <td><input name=\"RadixH2\" value=\"%d\"> <td> max %d min %d\n",
        require_radix_h2, require_radix_h2, RADIX_H2_MAX, 3);
    fdprintf(ofd, "<tr><td>RadixH3  <td align=\"right\"> %8d <td>  <td><input name=\"RadixH3\" value=\"%d\"> <td> max %d min %d\n",
        require_radix_h3, require_radix_h3, RADIX_H3_MAX, 3);
    fdprintf(ofd, "<tr><td>DiskSweepInterval  <td align=\"right\"> %8d <td>  <td><input name=\"DiskSweepInterval\" value=\"%d\"> <td> max %d min %d\n",
        require_disk_sweep_interval, require_disk_sweep_interval, 3600000, 100);
    fdprintf(ofd, "<tr><td>NFragObject  <td align=\"right\"> %8d <td>  <td><input name=\"NFragObject\" value=\"%d\"> <td> max %d min %d\n",
        require_nfrag_obj, require_nfrag_obj, HARD_NFRAG_OBJ, 1);
    fdprintf(ofd, "<tr><td>NFragNetWrite  <td align=\"right\"> %8d <td>  <td><input name=\"NFragNetWrite\" value=\"%d\"> <td> max %d min %d\n",
        require_nfrag_nwrite, require_nfrag_nwrite, HARD_NFRAG_NWRITE, 1);
    fdprintf(ofd, "<tr><td>NFragNetRead  <td align=\"right\"> %8d <td>  <td><input name=\"NFragNetRead\" value=\"%d\"> <td> max %d min %d\n",
        require_nfrag_nread, require_nfrag_nread, HARD_NFRAG_NREAD, 1);
    fdprintf(ofd, "<tr><td>NFragDiskWrite  <td align=\"right\"> %8d <td>  <td><input name=\"NFragDiskWrite\" value=\"%d\"> <td> max %d min %d\n",
        require_nfrag_dwrite, require_nfrag_dwrite, HARD_NFRAG_DWRITE, 1);
    fdprintf(ofd, "<tr><td>NFragDiskRead  <td align=\"right\"> %8d <td>  <td><input name=\"NFragDiskRead\" value=\"%d\"> <td> max %d min %d\n",
        require_nfrag_dread, require_nfrag_dread, HARD_NFRAG_DREAD, 1);
    fdprintf(ofd, "<tr><td>NIndexDiskRead  <td align=\"right\"> %8d <td>  <td><input name=\"NIndexDiskRead\" value=\"%d\"> <td> max %d min %d\n",
        require_nindex_dread, require_nindex_dread, HARD_NINDEX_DREAD, 1);
    fdprintf(ofd, "<tr><td>NIndexBurstDiskRead  <td align=\"right\"> %8d <td>  <td><input name=\"NIndexBurstDiskRead\" value=\"%d\"> <td> max %d min %d\n",
        require_nindex_bdread, require_nindex_bdread, HARD_NINDEX_BDREAD, 1);
    fdprintf(ofd, "<tr><td>FragHHwater  <td align=\"right\"> %8d <td>  <td><input name=\"FragHHwater\" value=\"%d\"> <td> max %d min %d\n",
        require_frag_hhwater, require_frag_hhwater, 100, 1);
    fdprintf(ofd, "<tr><td>FragHwater  <td align=\"right\"> %8d <td>  <td><input name=\"FragHwater\" value=\"%d\"> <td> max %d min %d\n",
        require_frag_hwater, require_frag_hwater, 100, 1);
    fdprintf(ofd, "<tr><td>FragLwater  <td align=\"right\"> %8d <td>  <td><input name=\"FragLwater\" value=\"%d\"> <td> max %d min %d\n",
        require_frag_lwater, require_frag_lwater, 100, 1);
    fdprintf(ofd, "<tr><td>EntryHHwater  <td align=\"right\"> %8d <td>  <td><input name=\"EntryHHwater\" value=\"%d\"> <td> max %d min %d\n",
        require_entry_hhwater, require_entry_hhwater, 100, 1);
    fdprintf(ofd, "<tr><td>EntryHwater  <td align=\"right\"> %8d <td>  <td><input name=\"EntryHwater\" value=\"%d\"> <td> max %d min %d\n",
        require_entry_hwater, require_entry_hwater, 100, 1);
    fdprintf(ofd, "<tr><td>EntryLwater  <td align=\"right\"> %8d <td>  <td><input name=\"EntryLwater\" value=\"%d\"> <td> max %d min %d\n",
        require_entry_lwater, require_entry_lwater, 100, 1);
    fdprintf(ofd, "<tr><td>NumDiskSweeper  <td align=\"right\"> %8d <td>  <td><input name=\"NumDiskSweeper\" value=\"%d\"> <td> max %d min %d\n",
        require_n_bsweeper, require_n_bsweeper, 32, 1);
    fdprintf(ofd, "<tr><td>NumMemorySweeper  <td align=\"right\"> %8d <td>  <td><input name=\"NumMemorySweeper\" value=\"%d\"> <td> max %d min %d\n",
        require_n_msweeper, require_n_msweeper, 32, 1);
    fdprintf(ofd, "<tr><td>NumSwapper  <td align=\"right\"> %8d <td>  <td><input name=\"NumSwapper\" value=\"%d\"> <td> max %d min %d\n",
        require_n_fswapper, require_n_fswapper, N_SWAPPER_MAX, 1);
    fdprintf(ofd, "<tr><td>TotalMemorySize  <td align=\"right\"> %8d <td> M (x 1048576) <td><input name=\"TotalMemorySize\" value=\"%d\"> <td> max %d min %d\n",
        require_total_memory_size/(1048576), require_total_memory_size/(1048576), 4096, 8);
    fdprintf(ofd, "<tr><td>NumBucket  <td align=\"right\"> %8d <td>  <td><input name=\"NumBucket\" value=\"%d\"> <td> max %d min %d\n",
        require_nbucket, require_nbucket, N_BUCKET_MAX, 1);
    fdprintf(ofd, "<tr><td>NEntryBucket  <td align=\"right\"> %8d <td>  <td><input name=\"NEntryBucket\" value=\"%d\"> <td> max %d min %d\n",
        require_nentry_per_bucket, require_nentry_per_bucket, 1024, 4);
    fdprintf(ofd, "<tr><td>LevelBucketSkipList  <td align=\"right\"> %8d <td>  <td><input name=\"LevelBucketSkipList\" value=\"%d\"> <td> max %d min %d\n",
        require_level_bucket_skiplist, require_level_bucket_skiplist, 64, 2);
    fdprintf(ofd, "<tr><td>FragSize  <td align=\"right\"> %8d <td>  <td><input name=\"FragSize\" value=\"%d\"> <td> max %d min %d\n",
        require_frag_size, require_frag_size, 8192, 128);
    fdprintf(ofd, "<tr><td>NumPrefetchReferedPages  <td align=\"right\"> %8d <td>  <td><input name=\"NumPrefetchReferedPages\" value=\"%d\"> <td> max %d min %d\n",
        num_prefetch_referedpages, num_prefetch_referedpages, 64, 0);
    fdprintf(ofd, "<tr><td>NumPrefetchIncludedObjects  <td align=\"right\"> %8d <td>  <td><input name=\"NumPrefetchIncludedObjects\" value=\"%d\"> <td> max %d min %d\n",
        num_prefetch_includedobjs, num_prefetch_includedobjs, 128, 0);
    fdprintf(ofd, "<tr><td>ThruFactor  <td align=\"right\"> %8d <td>  <td><input name=\"ThruFactor\" value=\"%d\"> <td> max %d min %d\n",
        require_thru_factor, require_thru_factor, 100, -1);
    fdprintf(ofd,
"\n\
\n\
</table>\n\
\n");

    fdprintf(ofd,
"<input name=\"send\" type=\"submit\" value=\"APPLY\">\n\
<input name=\"clear\" type=\"reset\" value=\"CLEAR\">\n");


    fdprintf(ofd, "</form>\n");

}


void
ShowProgramStatus(FILE *outport)
{
    fprintf(outport, "KOTETU [%s]\n", KOTETU_VER);

    fflush(outport);
}

void
ShowUsage()
{
    FixGlobals();   /* check in temporary */
    ShowProgramStatus(stdout);
    ShowGlobals(stdout);
}

#if 0
int
SetProcessComment(char *comment)
{
    char line[STRING_SIZE];

    sprintf(line, "wcol : %s", comment);
    return SetProcessState(line);
}
#endif


/*
 * ReadCommandlineOption
 */
void ReadCommandlineOptions(int argc, char **argv)
{
    int flag;

    extern char *optarg;
    extern int  optind;

    while((flag = getopt(argc, argv,
        "hvf:l:tPp:R:n:H:C:c:D:M:W:U:r:i:T:")) != -1) {

        switch(flag) {
        case 'h':
            ShowUsage();
            exit(KOTETU_EXIT_NORMAL);
            break;
        case 'v':
            ShowProgramStatus(stdout);
            exit(KOTETU_EXIT_NORMAL);
            break;
        case 'f':
            strcpy(config_filename, optarg);
            ReadConfigFile(config_filename);
            break;
        case 'l':
            log_filename = strdup(optarg);
            break;
        case 't':
            trace_flag = 1 - trace_flag;
            break;
        case 'P':
            prefetch_flag = 1 - prefetch_flag;
            break;
        case 'p':
            service_portno = (int) atoi(optarg);
            break;
        case 'R':
            report_hook_hostname = strdup(optarg);
            break;
        case 'n':
            require_nsess_total = (int) atoi(optarg);
            break;
        case 'H':
            require_n_handler = (int) atoi(optarg);
            break;
        case 'C':
            require_nsess_handler = (int) atoi(optarg);
            break;
        case 'c':
            cache_dir_prefix = strdup(optarg);
            break;
        case 'D':
            require_n_bsweeper = (int) atoi(optarg);
            break;
        case 'M':
            require_n_msweeper = (int) atoi(optarg);
            break;
        case 'W':
            require_n_fswapper = (int) atoi(optarg);
            break;
        case 'U':
            require_total_memory_size = (off_t) atoi(optarg) * 1048576;
            break;
        case 'r':
            num_prefetch_referedpages = (int) atoi(optarg);
            break;
        case 'i':
            num_prefetch_includedobjs = (int) atoi(optarg);
            break;
        case 'T':
            require_thru_factor = (int) atoi(optarg);
            break;
        }
    }
}



