/*
 * KOTETU: the prefetching WWW proxy server
 *
 * proxy_main.c : main body for proxy server.
 *
 *      by k-chinen@is.aist-nara.ac.jp, 1999-2001.
 *
 * $Id$
 */

/* global configurations */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


/* debug flags */

/*
#undef CHECK_BKMUTEX
*/

/* public headers */

#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <pthread.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#ifdef HAVE_POLL_H
#include <poll.h>
#include <sys/poll.h>
#endif
#ifdef HAVE_SELECT_H
#include <sys/select.h>
#endif
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif
#include <strings.h>
#include <limits.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <assert.h>
#include <errno.h>


/* private headers */

#include "range.h"
#include "msg.h"
#include "trace.h"
#include "mem.h"
#include "socket.h"
#include "sigs.h"
#include "sim_stat.h"
#include "sim_hist.h"
#include "misc.h"
#include "http.h"
#include "ocache.h"
#include "sess.h"
#include "squeue.h"
#include "parse.h"
#include "mlist.h"
#include "reason.h"




TRACE_REGIST_NIC("swapper");


#ifndef REAPER_NTRY
#define REAPER_NTRY (5)
#endif

extern pthread_mutex_t config_lock;

extern pthread_mutex_t stat_lock;
extern time_t boot_date;
extern time_t service_start_date;

extern volatile long long swapout_req_serial;


/* ocache_mem.c */
extern volatile int compaction_count;
extern volatile int msweeper_count;

extern sim_stat_rec msweep_time;
extern sim_stat_rec msweep_time_short;
extern sim_stat_rec compaction_time;

extern volatile int swfactor;

extern pthread_t th_swapper[N_SWAPPER_MAX];
int swapper_focus_bk[N_SWAPPER_MAX];
int swapper_focus_oid[N_SWAPPER_MAX];
int swapper_phase[N_SWAPPER_MAX];


extern int handler_phase[N_HANDLER_MAX];


extern int n_listener, n_handlers, n_bsweepers, n_swappers, n_monitors;

extern sess_queue_t *think_queue;

extern sess_queue_t X_swap_queue[N_SWAPPER_MAX];
extern sess_queue_t *swap_queue[N_SWAPPER_MAX];
extern volatile long long swap_op_th_count[N_SWAPPER_MAX][6];
extern volatile long long swap_op_all_count[6];
extern volatile long long swap_op_repush_th_count[N_SWAPPER_MAX];
extern volatile long long swap_op_release_th_count[N_SWAPPER_MAX];
extern volatile long long swap_op_retry_th_count[N_SWAPPER_MAX];

extern sim_hist_rec swapper_swo_hist[N_SWAPPER_MAX];
extern sim_hist_rec swapper_swi_hist[N_SWAPPER_MAX];


#ifdef CHECK_QUEUE

#define push_sess(q,c) \
    RawTrace("push sess# %d, %p to %p : %d/%d : %s %5d\n",\
        c->sid, c, q, q->count, q->cmax, __FILE__ , __LINE__), sq_push(q,c)
#define pull_sess(q) \
    RawTrace("pull from %p : %s %5d\n", q, __FILE__ , __LINE__), sq_pull(q)

#else
#define push_sess(q,c) sq_push(q,c)
#define pull_sess(q) sq_pull(q)
#endif
#define pull_sess_nonblock(q) sq_pull_nonblock(q)


extern int require_nsess_handler;


volatile int swapper_turn_all_time = 50000;
volatile int swapper_turn_each_time = 50000;



volatile int reaper_requested=0, reaper_called=0;


#include "reaper.c"

int
be_miss(skcache_entry_t *stored_pos, sess_t *sessp)
{
    stored_pos->owner       = sessp;
    stored_pos->oid         = sessp->sid;
    stored_pos->issue_time  = sessp->last_action;

if(stored_pos->issue_time<=0) {
Trace("no3 stored_pos->issue_time %d, sessp->last_action %d\n",
stored_pos->issue_time, sessp->last_action);
}

    stored_pos->totallength = (off_t) 0L;
    stored_pos->m_recorded = NULL;
    stored_pos->d_recorded = 0;

    return 0;
}


static
int
Smain(int thno, bucket_t *bk, sess_t *sessp)
{
    int chk;
    time_t now;
    int op;
    int rnum;
    struct timeval swo_begin, swo_end;
    struct timeval swi_begin, swi_end;
    int sig_pos;
    skcache_entry_t *stored_pos;

#if 0
    Trace("swapper# %d, bk# %d, state %d\n",
        thno, sessp->fhash, sessp->state);
#endif

    swapper_phase[thno] = 7;
    stored_pos = sessp->cache_entry;

    switch(sessp->state) {

    /***
     *** SWAPIN 
     ***/
    case STATE_WAIT_SWAPIN:

        swapper_phase[thno] = 101;
#if 0
        LOCK_BK(bk);
        swapper_phase[thno] = 102;
#endif

#if 0
chk = sk_entity_check(bk, __FILE__, __LINE__);
#endif

        gettimeofday(&swi_begin, NULL);
        chk = skcache_entry_swapin(bk, stored_pos);
        gettimeofday(&swi_end, NULL);
        swapper_phase[thno] = 103;
        if(chk) {
            if(chk==XSK_SWAP_MISS) {
#if 0
                Trace("swapper# %d, swapin() return MISS for sess# %d\n",
                    thno, sessp->sid);
#endif
            }
            else {
                if(chk>0) {
                    /* nothing */
                }
                else {
                Error("swapper# %d, fail swapin bk# %d, sess# %d <%d>",
                    thno, bk->bid, sessp->sid, chk);
                }
            }

            /* miss. of cause error mean miss.
               entity is set as new-one and throw back to handlers
             */

            swapper_phase[thno] = 111;
            MUTEX_LOCK(&stat_lock);
            miss_reply_count++;
            MUTEX_UNLOCK(&stat_lock);
            swapper_phase[thno] = 112;

            be_miss(stored_pos, sessp);
            sessp->state = STATE_QUERY_ADDR;
        }
        else {
#if 0
            Trace("swapper# %d, success swapin sess# %d,\n",
                thno, sessp->sid);
#endif
            swapper_focus_oid[thno] = stored_pos->oid;

#if 1
            /* XXX */
            if(stored_pos->ref!=0) {
                Error(
"swapper# %d, swapined object oid# %d, is too much refrenced (ref %d)",
                    thno, stored_pos->oid, stored_pos->ref);
            }
            stored_pos->ref++;
#endif

            swapper_phase[thno] = 121;
            MUTEX_LOCK(&stat_lock);
            hit_reply_count++;
            MUTEX_UNLOCK(&stat_lock);
            swapper_phase[thno] = 122;

            sessp->state = STATE_SEND_CACHE;
            sessp->ishit = SESS_HIT;

        }

        sim_hist_add_value(&swapper_swi_hist[thno], 
            KOTETU_SUB_TIME(swi_end, swi_begin));

#if 0
chk = sk_entity_check(bk, __FILE__, __LINE__);
#endif
#if 0
        UNLOCK_BK(bk);
#endif

        if(stored_pos->issue_time<=0) {
Trace("Swaper: sess# %d, state %s(%d) <addr %p> ignore issue_time\n",
            sessp->sid, MSG_STATE_[sessp->state], sessp->state, stored_pos);
        }


        /* throw this to thinkQ */
        swapper_phase[thno] = 171;
        push_sess(think_queue, sessp);
        swapper_phase[thno] = 172;

        sessp = NULL;

        swapper_focus_oid[thno] = -1;

        swap_op_repush_th_count[thno]++;

        break;
    
    case STATE_WAIT_SWAPOUT:
    case STATE_WAIT_MEMORY_GC:
    case STATE_WAIT_DISK_GC:
    default:
        Trace("swapper# %d, got %s(%d) session sess# %d,",
            thno, MSG_STATE_[sessp->state], sessp->state,
            sessp->sid);
        break;

    }

    if(sessp) {
        swapper_phase[thno] = 901;
        sq_be_free(sessp);  
        swapper_phase[thno] = 902;

        swap_op_release_th_count[thno]++;
    }
    swapper_phase[thno] = -7;
}


#ifndef SW_SCHUNK_LOW_HARD  
#define SW_SCHUNK_LOW_HARD      (16)
#endif
#ifndef SW_SCHUNK_HIGH_HARD
#define SW_SCHUNK_HIGH_HARD     (64)
#endif
#ifndef SW_TURNTIME_MAX_HARD
/*
#define SW_TURNTIME_MAX_HARD    (4096)
*/
#define SW_TURNTIME_MAX_HARD    (3*1024/2)
#endif
#ifndef SW_BKBLUK_MAX_HARD
#define SW_BKBLUK_MAX_HARD      (8)
#endif

void*
Swapper(void* dummy)
{
    int thno;
    sess_t *sessp;
#if 0
    skcache_entry_t *stored_pos;
#endif
    bucket_t *bk;
    bucket_t *last_bk;
    int chk;
    time_t now;
    int op;
    int rnum;
    int bkmask[RADIX_H1_MAX];
    int qn;
    int n_repeat;
    sess_t *sess_chunk[NSESS_HANDLER_MAX];
    int i;
    int sess_max, sess_count;

    struct timeval lbegin, lend;
    time_t elap;
    time_t takes;

    int p_prev, p_back;
    int last_prev, last_back;



    thno = (int)dummy;
    rnum = 0;

#ifdef SINGLE_SWAP_QUEUE
    qn = 0;
#else
    qn = thno/swfactor;
#endif


#if 0
    Trace("Swapper th# %d %#x for disk# %d\n",
        pthread_self(), pthread_self(), thno);
#endif

#if 0
    {
        int i;
        char tmp[L_PATH];
        char line[10*RADIX_H1_MAX];

        line[0] = '\0';
        for(i=0;i<RADIX_H1_MAX;i++) {
            bkmask[i] = -1;
        }
        for(i=0;i<radix_h1;i++) {
            if(SWSOLV(i)==thno) {
                bkmask[i] = 1;
                sprintf(tmp, "bk# %d, ", i);
                strcat(line, tmp);
            }
            else {
                bkmask[i] = 0;
            }
        }
#if 0
        Trace("Swapper %d %#x for disk# %d; %s \n",
            pthread_self(), pthread_self(), thno, line);
#endif
    }
#endif

    last_bk = NULL;
    sess_max = 32; 
    
    last_prev = 32;
    last_back = 0;


    while(1) {
        swapper_focus_bk[thno] = -1;
        swapper_focus_oid[thno] = -1;
        swapper_phase[thno] = -1;

        /* XXX */
        MUTEX_LOCK(&config_lock);

#if 1
        /* static */
        sess_max = 16;
#endif

#if 0
        /* static */
        sess_max = 32;
#endif

#if 0
        /* dynamic */
        if(last_back>0) {
            sess_max = MAX(SW_SCHUNK_LOW_HARD, last_prev+1);
        }
        else {
            sess_max = MIN(SW_SCHUNK_HIGH_HARD, sess_max+1);
        }
#endif

        swapper_turn_all_time = SW_TURNTIME_MAX_HARD;
        swapper_turn_each_time = 200;
        MUTEX_UNLOCK(&config_lock);

        for(i=0;i<sess_max;i++) {
            sess_chunk[i] = NULL;
        }

        sess_count = sq_pullchunk(swap_queue[qn], sess_chunk, sess_max);
        if(sess_count<=0) {
            continue;
        }
#if 0
        Trace("swapper# %d, got %d sessions (max %d)\n",
            thno, sess_count, sess_max);
#endif


        last_bk = NULL;
        p_prev = p_back = 0;

        gettimeofday(&lbegin, NULL);
        for(i=0;i<sess_count;i++) {
            p_prev++;
            sessp = sess_chunk[i];

            rnum++;
            if(sessp==NULL) {
                op = 0;
            }
            else {
                switch(sessp->state) {
                case STATE_WAIT_SWAPIN:     op = 1; break;
                case STATE_WAIT_SWAPOUT:    op = 2; break;
                case STATE_WAIT_DISK_GC:    op = 3; break;
                case STATE_WAIT_MEMORY_GC:  op = 4; break;
                default:                    op = 5; break;
                }
            }

#if 0
            Trace("swapper# %d, q# %d, op %d,\n", thno, thno/swfactor, op);
#endif

            swap_op_th_count[thno][op]++;
            MUTEX_LOCK(&stat_lock);
            swap_op_all_count[op]++;
            MUTEX_UNLOCK(&stat_lock);

            if(sessp==NULL)
                continue;

            bk = &skcache_bucket_base[sessp->fhash];

            swapper_focus_bk[thno] = bk->bid;
            swapper_phase[thno] = 3;

            swapper_phase[thno] = 601;
            LOCK_BK(bk);
            swapper_phase[thno] = 602;

            chk = Smain(thno, bk, sessp); 

            swapper_phase[thno] = 603;
            UNLOCK_BK(bk);
            swapper_phase[thno] = 604;

            gettimeofday(&lend, NULL);
            takes = KOTETU_SUB_TIME(lend, lbegin);
            if(takes > swapper_turn_all_time) {
                Trace("swapper# %d, process %d/%d, takes %d within %d\n",
                    thno, i+1, sess_count, takes, swapper_turn_all_time);
                i++;
                swapper_phase[thno] = 607;
                break;
            }
        }

        swapper_focus_bk[thno] = -1;

        for(;i<sess_count;i++) {
            swapper_phase[thno] = 913;
#if 0
            Trace("swapper# %d, th %08x, i %d, sendback\n",
                thno, pthread_self(), i);
#endif
            p_back++;

            sessp = sess_chunk[i];
            if(sessp==NULL) {
                continue;
            }

#if 0
            /* retry */
            swap_op_retry_th_count[thno]++;
#if 0
            push_sess(swap_queue[qn], sessp);
#endif
            sq_pullback(swap_queue[qn], sessp);

#else
            /* give up */

            bk = &skcache_bucket_base[sessp->fhash];

            swapper_phase[thno] = 701;
            LOCK_BK(bk);
            swapper_phase[thno] = 702;

            swapper_phase[thno] = 711;
            MUTEX_LOCK(&stat_lock);
            miss_reply_count++;
            MUTEX_UNLOCK(&stat_lock);
            swapper_phase[thno] = 712;

            be_miss(sessp->cache_entry, sessp);
            sessp->state = STATE_QUERY_ADDR;

            UNLOCK_BK(bk);
            
            /* throw this to thinkQ */
            swapper_phase[thno] = 771;
            push_sess(think_queue, sessp);
            swapper_phase[thno] = 772;

            sessp = NULL;

#endif
        }

        gettimeofday(&lend, NULL);
        elap = lend.tv_sec - boot_date;
#if 0
        if(thno==0) {
            Trace("swapper# %d, th %08x, q# %d, t %d, got %d %d left %d;  prev %d, back %d, takes %d/%d msec\n",
                thno, pthread_self(), qn, 
                elap,
                sess_count, sess_max,
                swap_queue[qn]->count,
                p_prev, p_back,
                KOTETU_SUB_TIME(lend, lbegin), swapper_turn_all_time);
        }
#endif
        last_prev = p_prev;
        last_back = p_back;

        swapper_phase[thno] = 915;
    }

    Error("swapper# %d, meltdown", thno);

    return;
}

void
swapper_init()
{
    TRACE_RECEIPT;
}

